fn find_min(heights: &[u32]) -> (usize, &u32) {
    heights.iter().enumerate().max_by_key(|(_, &x)| x).unwrap_or((0, &0))
}

///                      #            #
///                                   #
///                      ##           ##
///                      #            ##
///                                   ##             #          ##  #
///       #              #            ##             #          ##  #
///   #   ## #   rotate  ###  extend  ###  truncate  ##   wall  ### #
/// # ## ######    ->    ##     ->    ###     ->     ##    ->   ### #
///                      #            ###            ##         ### #
///                      ##           ###            ##         ### #
///                      #            ###
///
fn rotate_right(heights: &[u32]) -> Vec<u32> {
    let mut new = Vec::new();
    let (min, &min_elem) = find_min(heights);

    'outer: for i in 0..min_elem {
        for (j, &height) in heights.iter().enumerate() {
            if height > i {
                new.push((min - j) as u32);
                continue 'outer;
            }
        }
    }

    // wall
    new.insert(0, min as u32);
    new.push(min as u32);

    new
}

///                        #            #
///                       ##           ##
///                        #           ##
///                       ##           ##
///                      ###          ###
///       #                #          ###              #        #  ##
///   #   ## #   rotate       extend  ###  truncate   ##  wall  # ###
/// # ## ######    ->      #    ->    ###     ->      ##   ->   # ###
///                       ##          ###             ##        # ###
///                                   ###
///                        #          ###
///
fn rotate_left(heights: &[u32]) -> Vec<u32> {
    let mut new = Vec::new();
    let (min, &min_elem) = find_min(heights);
    let min = heights.len() - min - 1;

    'outer: for i in (0..min_elem).rev() {
        for (j, &height) in heights.iter().rev().enumerate() {
            if height > i {
                new.push((min - j) as u32);
                continue 'outer;
            }
        }
        break;
    }

    // wall
    new.insert(0, min as u32);
    new.push(min as u32);

    new
}

pub fn entry(height: &[u32]) -> u32 {
    if height.is_empty() || height.iter().all(|&x| x == height[0]) {
        return 0;
    }
    let s1 = entry(&rotate_right(height));
    let s2 = entry(&rotate_left(height));
    height.len() as u32 * height.iter().copied().max().unwrap_or(0) - height.iter().sum::<u32>() - s1 - s2
}
