import itertools
import functools
import operator


def vectorize(f):
    def g(*args):
        match args:
            case [*x], [*y]:
                return [*map(f, x, y)]
            case [*x], y:
                return [f(a, y) for a in x]
            case x, [*y]:
                return [f(x, b) for b in y]
            case x, y:
                return f(x, y)

    return g


def evaluate(ast, input):
    if ast == "a":
        return input
    if callable(ast):
        return ast
    acc = evaluate(ast[-1], input)
    i = len(ast) - 2
    while i >= 0:
        assert callable(f := ast[i])
        if i > 0 and not callable(lhs := evaluate(ast[i - 1], input)):
            acc = f(lhs, acc)
            i -= 2
        else:
            acc = f(acc)
            i -= 1
    return acc


builtins = {
    "+": vectorize(operator.add),
    "-": vectorize(operator.sub),
    "⌊": vectorize(min),
    "⌈": vectorize(max),
    "⌽": lambda x: list(reversed(x)),
}


def run(code, input):
    stack = [[]]
    for c in code:
        if c == "(":
            stack.append([])
        elif c == ")":
            top = stack.pop()
            stack[-1].append(top)
        elif c == "/":
            f = stack[-1].pop()
            stack[-1].append(lambda x, f=f: functools.reduce(f, x))
        elif c == "\\":
            f = stack[-1].pop()
            stack[-1].append(lambda x, f=f: list(itertools.accumulate(x, f)))
        else:
            stack[-1].append(builtins.get(c, c))
    return evaluate(stack.pop(), input)


entry = functools.partial(run, r"+/((⌈\a)⌊⌽⌈\⌽a)-a")

print(entry([0, 1, 0, 2, 1, 0, 1, 3, 2, 1, 2, 1]))
